from typing import Iterable

from pygame.color import Color
from pygame.surface import Surface

from robingame.image import brighten_color
from robingame.input import GamecubeController
from robingame.objects import Entity, Group, Particle


class SmashParticle(Particle):
    radius = 20
    decay = 2
    color = Color("red")
    gravity = 0
    friction = 0
    blit_flag = 0


class GamecubeControllerVisualizer(Entity):
    input: GamecubeController

    def __init__(self, x, y, input: GamecubeController, groups: Iterable = ()):
        super().__init__(*groups)
        self.x = x
        self.y = y
        self.input = input
        self.particles = Group()
        self.child_groups = [
            self.particles,
        ]

    def update(self):
        super().update()
        mapping = {
            self.input.LEFT: dict(x=10, y=10 + 50),
            self.input.RIGHT: dict(x=10 + 40, y=10 + 50),
            self.input.UP: dict(x=10 + 20, y=10 + 30),
            self.input.DOWN: dict(x=10 + 20, y=10 + 30 + 40),
            self.input.C_LEFT: dict(x=10 + 70, y=10 + 80 + 10, radius=10),
            self.input.C_RIGHT: dict(x=10 + 70 + 20, y=10 + 80 + 10, radius=10),
            self.input.C_UP: dict(x=10 + 70 + 10, y=10 + 80, radius=10),
            self.input.C_DOWN: dict(x=10 + 70 + 10, y=10 + 80 + 20, radius=10),
        }
        for input, kwargs in mapping.items():
            if input.is_smashed:
                self.particles.add(SmashParticle(**kwargs))

    def draw(self, surface, debug=False):
        background = Surface((160, 120))
        background.fill(Color("black"))

        # buttons
        self.draw_button(background, (20, 20), Color("cyan"), self.input.A, (100, 50))
        self.draw_button(background, (10, 10), Color("red"), self.input.B, (100, 80))
        self.draw_button(background, (20, 10), Color("gray"), self.input.Y, (100, 30))
        self.draw_button(background, (10, 20), Color("gray"), self.input.X, (130, 50))
        self.draw_button(background, (10, 10), Color("gray"), self.input.START, (65, 50))
        self.draw_button(background, (10, 10), Color("purple"), self.input.Z, (130, 30))
        self.draw_button(background, (7, 7), Color("gray"), self.input.D_PAD_UP, (40 + 8, 87 - 8))
        self.draw_button(background, (7, 7), Color("gray"), self.input.D_PAD_DOWN, (40 + 8, 87 + 8))
        self.draw_button(background, (7, 7), Color("gray"), self.input.D_PAD_LEFT, (40, 87))
        self.draw_button(background, (7, 7), Color("gray"), self.input.D_PAD_RIGHT, (40 + 16, 87))

        super().draw(background, debug)
        # sticks
        self.draw_joystick(
            background,
            (40, 40),
            Color("gray"),
            x_axis=(self.input.LEFT, self.input.RIGHT),
            y_axis=(self.input.UP, self.input.DOWN),
            xy=(0, 30),
        )
        self.draw_joystick(
            background,
            (20, 20),
            Color("orange"),
            x_axis=(self.input.C_LEFT, self.input.C_RIGHT),
            y_axis=(self.input.C_UP, self.input.C_DOWN),
            xy=(70, 80),
        )

        self.draw_trigger(
            background,
            (40, 20),
            Color("gray"),
            axis=self.input.R_AXIS,
            button=self.input.R,
            xy=(100, 0),
        )
        self.draw_trigger(
            background,
            (40, 20),
            Color("gray"),
            axis=self.input.L_AXIS,
            button=self.input.L,
            xy=(0, 0),
        )
        rect = background.get_rect()
        rect.topleft = self.x, self.y
        surface.blit(background, rect)

    def draw_trigger(self, surface, size, color, axis, button, xy):
        dx, dy = xy

        bbox = Surface(size)
        bbox.fill(color)
        bbox_rect = bbox.get_rect()
        bbox_rect.topleft = (dx + 10, dy + 10)
        surface.blit(bbox, bbox_rect)

        btn = Surface((size[0], size[1] - (size[1] * 0.8 * axis.value)))
        btn.fill(Color("red") if button.is_down else brighten_color(color, 100))
        btn_rect = btn.get_rect()
        btn_rect.centerx = bbox_rect.centerx
        btn_rect.bottom = bbox_rect.bottom
        surface.blit(btn, btn_rect)

    def draw_joystick(self, surface, size, color, x_axis, y_axis, xy):
        left, right = x_axis
        up, down = y_axis
        x_axis = right - left
        y_axis = down - up
        bbox = Surface(size)
        bbox.fill(color)
        bbox_rect = bbox.get_rect()
        dx, dy = xy
        bbox_rect.topleft = (dx + 10, dy + 10)
        surface.blit(bbox, bbox_rect)

        stick = Surface(tuple(s // 2 for s in size))
        is_smashed = any(input.is_smashed for input in (left, right, up, down))
        stick.fill(Color("red") if is_smashed else brighten_color(color, 100))
        rect = stick.get_rect()
        rect.center = (
            bbox_rect.centerx + x_axis * size[0] // 2,
            bbox_rect.centery + y_axis * size[1] // 2,
        )
        surface.blit(stick, rect)

    def draw_button(self, surface, size, color, input, xy):
        # B button
        btn = Surface(size)
        btn.fill(Color("white") if input.is_down else color)
        rect = btn.get_rect()
        dx, dy = xy
        rect.topleft = (dx + 10, dy + 10)
        surface.blit(btn, rect)


if __name__ == "__main__":

    class GCVizGame(Game):
        screen_color = Color("darkgray")

        def __init__(self):
            super().__init__()
            self.input = GamecubeController(controller_id=0)
            self.objects = Group()
            self.child_groups = [
                self.objects,
            ]
            self.visualizer = GamecubeControllerVisualizer(
                x=0,
                y=0,
                input=self.input,
                groups=[self.objects],
            )

        def read_inputs(self):
            super().read_inputs()
            self.input.read_new_inputs()

    GCVizGame().main()



import pytest
from pygame import Surface, Color

from robingame.image import (
    SpriteAnimation,
    load_spritesheet,
    load_image_sequence,
    load_image,
    relative_folder,
    brighten_color,
)

mocks = relative_folder(__file__, "mocks")


@pytest.mark.parametrize(
    "num_images, expected_len",
    [
        (None, 3),
        (2, 2),
    ],
)
def test_load_spritesheet(num_images, expected_len):
    filename = mocks / "123_spritesheet.png"
    images = load_spritesheet(filename=filename, image_size=(64, 64), num_images=num_images)
    assert len(images) == expected_len
    assert isinstance(images[0], Surface)


def test_load_spritesheet_not_found():
    with pytest.raises(FileNotFoundError) as e:
        load_spritesheet(filename="foo/bar.png", image_size=(1, 1))
    assert str(e.value) == "Couldn't find foo/bar.png"


def test_load_image_sequence_not_found():
    with pytest.raises(FileNotFoundError) as e:
        load_image_sequence(filename="foo/bar")
    assert str(e.value) == "Couldn't find foo/bar"


@pytest.mark.parametrize(
    "num_images, expected_len",
    [
        (None, 3),
        (2, 2),
    ],
)
def test_load_image_sequence(num_images, expected_len):
    filename = mocks / "123_series.png"
    images = load_image_sequence(filename=filename, num_images=num_images)
    assert len(images) == expected_len
    assert isinstance(images[0], Surface)


def test_can_instantiate_empty_spriteanimation():
    """images=None by default and it shouldn't try to flip/recolor/scale"""
    SpriteAnimation()


def test_spriteanimation_from_spritesheet():
    filename = mocks / "123_spritesheet.png"
    anim = SpriteAnimation.from_spritesheet(filename=filename, image_size=(64, 64))
    assert isinstance(anim, SpriteAnimation)
    assert len(anim.images) == 3
    assert isinstance(anim.images[0], Surface)


def test_spriteanimation_from_image_sequence():
    filename = mocks / "123_series.png"
    anim = SpriteAnimation.from_image_sequence(filename=filename)
    assert isinstance(anim, SpriteAnimation)
    assert len(anim.images) == 3
    assert isinstance(anim.images[0], Surface)


def test_spriteanimation_from_image():
    filename = mocks / "123_spritesheet.png"
    anim = SpriteAnimation.from_image(filename=filename)
    assert isinstance(anim, SpriteAnimation)
    assert len(anim.images) == 1
    assert isinstance(anim.images[0], Surface)


def test_spriteanimation_copy_methods():
    pxl = Surface((2, 2))
    pxl.fill(Color("red"))
    pxl.set_at((0, 0), Color("black"))

    anim = SpriteAnimation(images=[pxl])
    assert anim.images[0].get_at((0, 0)) == Color("black")
    assert anim.images[0].get_at((0, 1)) == Color("red")
    assert anim.images[0].get_at((1, 0)) == Color("red")
    assert anim.images[0].get_at((1, 1)) == Color("red")

    hor_flip = anim.flipped_copy(flip_x=True)
    assert hor_flip.images[0].get_at((0, 0)) == Color("red")
    assert hor_flip.images[0].get_at((0, 1)) == Color("red")
    assert hor_flip.images[0].get_at((1, 0)) == Color("black")
    assert hor_flip.images[0].get_at((1, 1)) == Color("red")

    ver_flip = anim.flipped_copy(flip_y=True)
    assert ver_flip.images[0].get_at((0, 0)) == Color("red")
    assert ver_flip.images[0].get_at((0, 1)) == Color("black")
    assert ver_flip.images[0].get_at((1, 0)) == Color("red")
    assert ver_flip.images[0].get_at((1, 1)) == Color("red")

    recolored = anim.recolored_copy({(0, 0, 0): Color("blue")})
    assert recolored.images[0].get_at((0, 0)) == Color("blue")
    assert recolored.images[0].get_at((0, 1)) == Color("red")
    assert recolored.images[0].get_at((1, 0)) == Color("red")
    assert recolored.images[0].get_at((1, 1)) == Color("red")

    scaled = anim.scaled_copy(scale=3)
    assert anim.images[0].get_rect().width == 2
    assert anim.images[0].get_rect().height == 2
    assert scaled.images[0].get_rect().width == 6
    assert scaled.images[0].get_rect().height == 6

    # original should be unchanged
    anim = SpriteAnimation(images=[pxl])
    assert anim.images[0].get_at((0, 0)) == Color("black")
    assert anim.images[0].get_at((0, 1)) == Color("red")
    assert anim.images[0].get_at((1, 0)) == Color("red")
    assert anim.images[0].get_at((1, 1)) == Color("red")


def test_load_image_with_per_pixel_transparency():
    filename = mocks / "per_pixel_alpha.png"
    image = load_image(filename.as_posix())

    # white and red pixels should have full alpha
    for red_pixel in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        assert image.get_at(red_pixel) == (255, 0, 0, 255)
    for white_pixel in [(2, 2), (2, 3), (3, 2), (3, 3)]:
        assert image.get_at(white_pixel) == (255, 255, 255, 255)

    # green pixels should have alpha = 100
    for green_pixel in [(2, 0), (2, 1), (3, 1), (1, 3), (1, 2), (0, 2)]:
        assert image.get_at(green_pixel) == (0, 255, 0, 100)

    # corner pixels should be fully transparent
    for corner_pixel in [(3, 0), (0, 3)]:
        assert image.get_at(corner_pixel) == (0, 0, 0, 0)


def test_load_image_with_global_transparency():
    filename = mocks / "global_alpha.png"
    image = load_image(filename.as_posix())

    # white and red and green pixels should have full alpha
    for red_pixel in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        assert image.get_at(red_pixel) == (255, 0, 0, 255)
    for white_pixel in [(2, 2), (2, 3), (3, 2), (3, 3)]:
        assert image.get_at(white_pixel) == (255, 255, 255, 255)
    for green_pixel in [(2, 0), (2, 1), (3, 1), (1, 3), (1, 2), (0, 2)]:
        assert image.get_at(green_pixel) == (0, 255, 0, 255)

    # corner pixels should be fully transparent
    for corner_pixel in [(3, 0), (0, 3)]:
        assert image.get_at(corner_pixel) == (0, 0, 0, 0)


@pytest.mark.parametrize(
    "amount, old_color, new_color",
    [
        (
            20,
            (0, 0, 0),
            (20, 20, 20),
        ),
        (
            20,
            Color(0, 0, 0),
            (20, 20, 20),
        ),
        (
            20,
            (250, 250, 250),
            (255, 255, 255),
        ),
        (
            20,
            (255, 255, 255),
            (255, 255, 255),
        ),
        (
            20,
            (0, 250, 255),
            (20, 255, 255),
        ),
        (
            20,
            (0, 0, 0, 0),
            (20, 20, 20, 0),
        ),
        (
            20,
            Color(0, 0, 0, 0),
            (20, 20, 20, 0),
        ),
        (
            -50,
            Color(0, 0, 0, 0),
            (0, 0, 0, 0),
        ),
        (
            -50,
            Color(0, 30, 55, 0),
            (0, 0, 5, 0),
        ),
    ],
)
def test_brighten_color(amount, old_color, new_color):
    assert brighten_color(old_color, amount=amount) == new_color


def test_subsurface():
    filename = mocks / "padded.png"
    image = load_image(filename.as_posix())
    assert image.get_width() == 16
    assert image.get_height() == 6

    x, y, w, h = image.get_bounding_rect()
    assert x == 7
    assert y == 2
    assert w == 2
    assert h == 2

    new = image.subsurface(image.get_bounding_rect())
    assert new.get_rect() == (0, 0, 2, 2)



import pytest

from robingame.utils import count_edges, SparseMatrix, unzip


@pytest.mark.parametrize(
    "input, expected_rising_edges, expected_falling_edges",
    [
        ([], 0, 0),
        ([0], 0, 0),
        ([1], 0, 0),
        ([0, 0], 0, 0),
        ([1, 1], 0, 0),
        ([0, 1], 1, 0),
        ([1, 0], 0, 1),
        ([1, 1, 1, 1, 1, 1], 0, 0),
        ([1, 0, 1, 1, 1, 1], 1, 1),
        ([1, 0, 1, 1, 0, 1], 2, 2),
        ([1, 0, 1, 0, 0, 1], 2, 2),
        ([0, 0, 0, 0, 0, 0, 0, 1], 1, 0),
        ([1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0], 0, 1),
        ([0, 0, 0, 0, 0.6, 0, 0, 0], 1, 1),
    ],
)
def test_count_edges(input, expected_rising_edges, expected_falling_edges):
    rising, falling = count_edges(input)
    assert rising == expected_rising_edges
    assert falling == expected_falling_edges


def test_unzip():
    a = (1, 2, 3, 4)
    b = (10, 20, 30, 40)
    c = (100, 200, 300, 400)
    zipped = list(zip(a, b, c))
    assert zipped == [
        (1, 10, 100),
        (2, 20, 200),
        (3, 30, 300),
        (4, 40, 400),
    ]
    assert unzip(zipped) == (a, b, c)


@pytest.mark.parametrize(
    "entries, expected_xlim, expected_ylim",
    [
        ([], (None, None), (None, None)),
        ([(2, 4)], (2, 2), (4, 4)),
        ([(2, 4), (3, 5)], (2, 3), (4, 5)),
        ([(10, 10), (-69, -420)], (-69, 10), (-420, 10)),
        ([(0, 0), (69, 420)], (0, 69), (0, 420)),
    ],
)
def test_sparse_matrix_xlim_ylim(entries, expected_xlim, expected_ylim):
    m = SparseMatrix({e: 1 for e in entries})
    assert m.xlim == expected_xlim
    assert m.ylim == expected_ylim


@pytest.mark.parametrize(
    "entries, expected_width, expected_height",
    [
        ([], 0, 0),
        ([(2, 4)], 1, 1),
        ([(2, 4), (3, 5)], 2, 2),
        ([(10, 10), (-69, -420)], 80, 431),
        ([(0, 0), (69, 420)], 70, 421),
    ],
)
def test_sparse_matrix_size_properties(entries, expected_width, expected_height):
    m = SparseMatrix({e: 1 for e in entries})
    assert m.width == expected_width
    assert m.height == expected_height
    assert m.size == (expected_width, expected_height)


def test_sparse_matrix_copy():
    m = SparseMatrix()
    m[(1, 1)] = True

    c = m.copy()
    assert isinstance(c, SparseMatrix)
    assert c[(1, 1)] is True



from unittest.mock import patch

import pygame
import pytest

from robingame.input import gamecube
from robingame.input.gamecube import (
    GamecubeController,
    ButtonInput,
    AxisInput,
)
from robingame.input.queue import InputQueue


@patch("pygame.joystick.Joystick")
@patch("robingame.input.gamecube.GamecubeControllerReader.get_values")
def test_gamecube_controller_basic(mock_get_values, mock_joystick):
    mock_get_values.return_value = (1, 0)  # A down, B not down
    pygame.init()

    controller = GamecubeController(controller_id=0)
    controller.read_new_inputs()
    assert controller[0] == (1, 0)
    assert controller.is_down(gamecube.A) == 1
    assert controller.is_down(gamecube.B) == 0
    assert controller.A.is_down == 1
    assert controller.B.is_down == 0


@patch("pygame.joystick.Joystick")
@patch("robingame.input.gamecube.GamecubeControllerReader.get_values")
def test_gamecube_controller_subclasses(mock_get_values, mock_joystick):
    mock_get_values.return_value = (1, 0)  # A down, B not down
    pygame.init()

    class Subclass(GamecubeController):
        A2 = ButtonInput(gamecube.A)

    class Subclass2(Subclass):
        A3 = ButtonInput(gamecube.A)

    class Subclass3(GamecubeController):
        B2 = ButtonInput(gamecube.B)

    subclass = Subclass(controller_id=0)
    subclass.read_new_inputs()
    assert subclass[0] == (1, 0)
    assert subclass.is_down(gamecube.A) == 1
    assert subclass.is_down(gamecube.B) == 0
    assert subclass.A.is_down == 1
    assert subclass.A2.is_down == 1
    with pytest.raises(AttributeError):
        assert subclass.A3.is_down == 1
    with pytest.raises(AttributeError):
        assert subclass.B2.is_down == 1
    assert subclass.B.is_down == 0

    subclass2 = Subclass2(controller_id=0)
    subclass2.read_new_inputs()
    assert subclass2[0] == (1, 0)
    assert subclass2.is_down(gamecube.A) == 1
    assert subclass2.is_down(gamecube.B) == 0
    assert subclass2.A.is_down == 1
    assert subclass2.A2.is_down == 1
    assert subclass2.A3.is_down == 1
    with pytest.raises(AttributeError):
        assert subclass.B2.is_down == 1
    assert subclass2.B.is_down == 0

    subclass3 = Subclass3(controller_id=0)
    subclass3.read_new_inputs()
    assert subclass3[0] == (1, 0)
    assert subclass3.is_down(gamecube.A) == 1
    assert subclass3.is_down(gamecube.B) == 0
    assert subclass3.A.is_down == 1
    with pytest.raises(AttributeError):
        assert subclass3.A2.is_down == 1
    with pytest.raises(AttributeError):
        assert subclass3.A3.is_down == 1
    assert subclass3.B.is_down == 0
    assert subclass3.B2.is_down == 0


@pytest.mark.parametrize(
    "input, expected_rising_edges, expected_falling_edges",
    [
        ([], 0, 0),  # no buffered values shouldn't crash things
        ([0, 0, 0], 0, 0),  # some values but less than the buffer
        ([0, 0, 1], 1, 0),
        ([1, 0, 0], 0, 1),
        ([0, 1, 0], 1, 1),
        ([0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0], 0, 0),  # too far in past
        ([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1], 1, 0),
        ([0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0], 1, 1),
        ([0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1], 2, 1),
    ],
)
def test_buffered_inputs(input, expected_rising_edges, expected_falling_edges):
    BUFFER_LENGTH = 5
    queue = InputQueue(queue_length=100)
    for value in input:
        queue.append([value])

    rising, falling = queue.buffered_inputs(key=0, buffer_length=BUFFER_LENGTH)

    assert rising == queue.buffered_presses(0, BUFFER_LENGTH) == expected_rising_edges
    assert falling == queue.buffered_releases(0, BUFFER_LENGTH) == expected_falling_edges


@pytest.mark.parametrize(
    "queue_contents, expected_value",
    [
        ([], False),  # empty queue
        ([0], False),  # only one entry; can't determine history
        ([1], False),  # only one entry; can't determine history
        ([0, 0.5], False),  # didn't reach the threshold
        ([0.5, 1], False),  # didn't start from low enough
        ([0, 0.3, 0.6, 0.8, 1], False),  # too slow
        ([0, 0.3, 0.6, 1], True),
        ([0, 0.5, 1], True),
        ([0, 1], True),
    ],
)
def test_axis_input_is_smashed(queue_contents, expected_value):
    queue = InputQueue(queue_length=100)
    for value in queue_contents:
        queue.append([value])
    axis = AxisInput(id=0, parent=queue)
    assert axis.is_smashed == expected_value



from robingame.gui import Button, Menu
from robingame.gui.menu import MyMenu
from robingame.image import SpriteAnimation, SpriteDict
from robingame.input import InputQueue, EventQueue, KeyboardInputQueue, GamecubeController
from robingame.objects import Game, Entity, PhysicalEntity, FpsTracker, Group, Particle
from robingame.text.font import Font, fonts


def test_imports():
    assert (
        Game,
        Entity,
        PhysicalEntity,
        Group,
        FpsTracker,
        Particle,
        Menu,
        Button,
        MyMenu,
        InputQueue,
        EventQueue,
        KeyboardInputQueue,
        GamecubeController,
        Font,
        fonts,
        SpriteAnimation,
    )



from pathlib import Path

from pygame.surface import Surface

from robingame.text.font import Font

TESTFONT = Path(__file__).parent.parent / "robingame/text/assets/test_font.png"
assert TESTFONT.exists()


def test_trim():
    font = Font(filename=TESTFONT, image_size=(16, 16), letters="A", trim=True)
    A = font.get("A")
    assert A.get_width() == 11  # trimmed width
    assert A.get_height() == 16  # original height


def test_render_respects_trimmed_character_size():
    font = Font(filename=TESTFONT, image_size=(16, 16), letters="A", trim=True)
    A = font.get("A")
    assert A.get_width() == 11  # trimmed width
    assert A.get_height() == 16  # original height

    x, y, w, h = A.get_bounding_rect()
    assert w == 11
    assert h == 13
    assert x == 0
    assert y == 2  # from the top of the image

    surf = Surface((100, 100)).convert_alpha()
    surf.fill((0, 0, 0, 0))
    assert surf.get_bounding_rect() == (0, 0, 0, 0)

    font.render(surf, "AAA")
    assert surf.get_bounding_rect() == (0, 2, w * 3, h)



from dataclasses import dataclass
from unittest.mock import patch

import pygame.event
from pygame.event import EventType, Event
from robingame.input.event import EventQueue


@patch("pygame.event.get")
def test_update(mock):
    assert EventQueue.events == []

    mock.return_value = ["foo"]
    EventQueue.update()
    assert EventQueue.events == ["foo"]

    mock.return_value = ["bar"]
    EventQueue.update()
    assert EventQueue.events == ["bar"]


def test_add_event_pygame_style():
    # in events.py we store an int representing the event type
    TEST_EVENT = pygame.event.custom_type()

    # event is created in the bowels of the code and we can pass any old stuff to the event __init__
    EventQueue.add(Event(TEST_EVENT, foo="foo", bar=69))

    # it shouldn't appear on the EventQueue.events list yet
    assert EventQueue.events == []

    # but after update, it should
    EventQueue.update()
    event = EventQueue.get(type=TEST_EVENT, foo="foo", bar=69)
    assert isinstance(event, EventType)

    # we trust that the event listener knows what attributes to look for
    assert event.foo == "foo"
    assert event.bar == 69
    assert event.type == TEST_EVENT


def test_add_custom_dataclass_event_is_converted_to_event():
    # in events.py
    @dataclass
    class MyCustomEvent:
        type = pygame.event.custom_type()
        foo: str
        bar: int = 0

    # somewhere in the code
    EventQueue.add(MyCustomEvent(foo="foo"))
    EventQueue.add(MyCustomEvent(foo="foo", bar=420))

    EventQueue.update()

    # dataclasses should be converted to Events by EventQueue
    assert len(EventQueue.filter(type=MyCustomEvent.type, foo="foo")) == 2
    result1 = EventQueue.get(type=MyCustomEvent.type, foo="foo", bar=0)
    assert isinstance(result1, EventType)
    assert result1.foo == "foo"
    assert result1.bar == 0
    result2 = EventQueue.get(type=MyCustomEvent.type, foo="foo", bar=420)
    assert isinstance(result2, EventType)
    assert result2.foo == "foo"
    assert result2.bar == 420



import pytest
from robingame.animation import ease_in, ease_out, ease_in_out


@pytest.mark.parametrize(
    "func",
    [
        ease_in,
        ease_out,
        ease_in_out,
    ],
)
def test_ease_functions(func):
    assert func(x=0, start=3, stop=10, num=50) == 3
    assert func(x=50, start=3, stop=10, num=50) == 10
    with pytest.raises(ValueError) as e:
        func(x=51, start=3, stop=10, num=50)

    assert str(e.value) == "x=51 is greater than num=50"



import pygame.draw
from pygame.color import Color
from pygame.surface import Surface

from robingame.objects import Game
from robingame.text.font import fonts

snippet = """
 1space
  2space
Ook    in 
de     spelling 
van    sommige andere talen wordt de umlaut gebruikt, bijvoorbeeld in verwante 
talen als het IJslands en het Zweeds, om een soortgelijk 

klankverschijnsel
weer
te;;;'''

geven, of in niet-verwante talen als het Fins, het Hongaars of het Turks, om dezelfde e-, eu- en u-klank weer te geven, die echter niet het resultaat van hetzelfde klankverschijnsel zijn.
"""


class FontTest(Game):
    window_width = 1500
    screen_color = (150, 150, 150)

    def draw(self, surface: Surface, debug: bool = False):
        super().draw(surface, debug)
        X = 500
        Y = 30
        WRAP = 500
        fonts.cellphone_white.render(
            surface,
            snippet,
            scale=2,
            wrap=WRAP,
            x=X,
            y=Y,
            align=-1,
        )
        pygame.draw.rect(surface, color=Color("red"), rect=(X, Y, WRAP, WRAP), width=1)


if __name__ == "__main__":
    FontTest().main()
